/**
 ** Copyright (c) 2001 LAAS/CNRS 
 **
 ** Matthieu Herrb - Fri Apr  6 2001
 **
 ** $Source: /cvs/openrobots/grh2/modules/camera/cameraImage.c,v $
 ** $Revision: 1.1 $
 ** $Date: 2006/01/19 16:45:31 $
 ** 
 **/
#ident "$Id: cameraImage.c,v 1.1 2006/01/19 16:45:31 matthieu Exp $"

#include <stdlib.h>
#include <tcl.h>
#include <tk.h>

#include <portLib.h>
#include <h2errorLib.h>

#include <auto/cameraPosterLib.h>

#define CAMERA_POSTER_VERSION "0.0"

static POSTER_ID cameraImagePosterId = NULL;

int 
cameraPosterMatchProc(Tcl_Obj *data, Tcl_Obj *formatString,
		      int *widthPtr, int *heightPtr, Tcl_Interp *interp)
{
    char *format = Tcl_GetString(formatString);
    char *posterName = Tcl_GetString(data);
    
    if (strcmp(format, "cameraPoster")) {
	/* Pas le bon format */
	return 0;
    }
    printf("camera poster name: %s\n", posterName);
    if (posterFind(posterName, &cameraImagePosterId) == ERROR) {
	/* poster inexistant */
	printf("erreur posterFind\n");
	return 0;
    }
    *widthPtr = CAMERA_NB_COL;
    *heightPtr = CAMERA_NB_LIG;
    printf("camera poster format %dx%d\n", *widthPtr, *heightPtr);
    return 1;
}

int 
cameraPosterReadProc(Tcl_Interp *interp, Tcl_Obj *data,
		     Tcl_Obj *formatString, Tk_PhotoHandle imageHandle,
		     int destX, int destY, int width, int height,
		     int srcX, int srcY)
{
    unsigned char *pixels;
    Tk_PhotoImageBlock b;

    if (cameraImagePosterId == NULL) {
	/* Poster inexistant */
	Tcl_SetResult(interp, "cameraPosterReadProc: poster Image NULL\n",
		      TCL_STATIC);
	return TCL_ERROR;
    }
    if (posterTake(cameraImagePosterId, POSTER_READ) == ERROR) {
	printf("posterTake error\n");
	return TCL_ERROR;
    }
    pixels = posterAddr(cameraImagePosterId);
    b.pixelPtr = pixels + srcY*CAMERA_NB_COL + srcX;
    b.width = CAMERA_NB_COL;
    b.height = CAMERA_NB_LIG;
    b.pitch = CAMERA_NB_COL;
    b.pixelSize = 1;
    b.offset[0] = 0;
    b.offset[1] = 0;
    b.offset[2] = 0;
    b.offset[3] = 0;
    
    Tk_PhotoPutBlock(imageHandle, &b, destX, destY, width, height);
    posterGive(cameraImagePosterId);
    return TCL_OK;
}

/**
 ** Cree un objet TCL a partir du poster de l'image compressee
 **
 ** Ensuite on peut utiliser le module tkjpeg de GDHE pour decompresser
 ** l'image et l'afficher
 ** 
 ** exemple:
 ** load /usr/local/robots/gdhe/lib/$env(MACHTYPE)-$env(OSTYPE)/tkjpeg.so
 ** image create photo cam -type jpeg \
 **          -data [cameraJpegPosterRead cameraJpegImage]]
 ** .c create image [expr [image width cam]/2] [expr [image height cam]/2] \
 **          -image cam 
 ** 
 **/
int
cameraJpegPosterRead(ClientData clientData, Tcl_Interp *interp,
		     int argc, Tcl_Obj *const argv[])
{
    char *posterName;
    POSTER_ID jpegPosterId;
    Tcl_Obj *result;

    /* Argument: nom du poster */
    if (argc != 2) {
	Tcl_SetResult(interp, "cameraJpegPosterRead: wrong # of args."
		      "Should be cameraJpegPosterRead <postername>",
		      TCL_STATIC);
	return ERROR;
    }
    posterName = Tcl_GetString(argv[1]);

    /* Recherche du poster */
    if (posterFind(posterName, &jpegPosterId) == ERROR) {
	Tcl_SetResult(interp, "cameraJpegPosterRead: poster not found",
		      TCL_STATIC);
	return TCL_ERROR;
    }
    if (posterTake(jpegPosterId, POSTER_READ) == ERROR) {
	Tcl_SetResult(interp, "cameraJpegPosterRead: poster Take error",
		      TCL_STATIC);
	return TCL_ERROR;
    }
    /* Copie du poster dans un nouvel objet TCL */
    /* XXX Comme on ne connait pas la taille on doit lire tout,
       XXX du coup on ne gagne rien !!! 
       XXX il faudrait lire l'entete jpeg pour connaitre la taille reelle 
       XXX a lire */
    result = Tcl_NewByteArrayObj(posterAddr(jpegPosterId), 
				 CAMERA_NB_COL*CAMERA_NB_LIG);
    posterGive(jpegPosterId);

    /* C'est tout */
    Tcl_SetObjResult(interp, result);
    return TCL_OK;
}
	
    
/**
 ** Init du module Tcl
 **/
int
Cameraimage_Init(Tcl_Interp *interp)
{
    int code;
    Tk_PhotoImageFormat cameraPosterFormat;

    cameraPosterFormat.name = "cameraPoster";
    cameraPosterFormat.fileMatchProc = NULL;
    cameraPosterFormat.stringMatchProc = cameraPosterMatchProc;
    cameraPosterFormat.fileReadProc = NULL;
    cameraPosterFormat.stringReadProc = cameraPosterReadProc;
    cameraPosterFormat.fileWriteProc = NULL;
    cameraPosterFormat.stringWriteProc = NULL;

    Tk_CreatePhotoImageFormat(&cameraPosterFormat);

    Tcl_CreateObjCommand(interp, "cameraJpegPosterRead", 
			 cameraJpegPosterRead, NULL, NULL);

    code = Tcl_PkgProvide(interp, "cameraPoster", CAMERA_POSTER_VERSION);
    if (code != TCL_OK) {
	return code;
    }
    return TCL_OK;
}


