/**
 ** Copyright (c) 1999 LAAS/CNRS 
 **
 ** Matthieu Herrb - Fri Sep 10 1999
 **
 ** $Source: /cvs/openrobots/grh2/modules/posters/locoPoster.c,v $
 ** $Revision: 1.1 $
 ** $Date: 2006/01/19 16:45:31 $
 ** 
 **/
#ident "$Id: locoPoster.c,v 1.1 2006/01/19 16:45:31 matthieu Exp $"

#include <stdlib.h>
#include <tcl.h>

#include <portLib.h>
#include <h2errorLib.h>

#include <tclPoster.h>

#include <auto/locoPosterLib.h>

int locoPoster_Init(Tcl_Interp *interp);

static TCL_POSTER_ENUM_VALUES locoMotionDirectionEnum[] = {
    { 0, "LOCO_FORWARD_DIRECTION" },
    { 1, "LOCO_BACKWARD_DIRECTION" }
};

static TCL_POSTER_DESC tclLocoPoster[] = {
    { "Robot.Position.x", TCL_POSTER_DOUBLE, 
      0, sizeof(double), NULL },
    { "Robot.Position.y", TCL_POSTER_DOUBLE,
      8, sizeof(double), NULL },
    { "Robot.Position.xC", TCL_POSTER_DOUBLE,
      16, sizeof(double), NULL },
    { "Robot.Position.yC", TCL_POSTER_DOUBLE,
      24, sizeof(double), NULL },
    { "Robot.Position.theta", TCL_POSTER_DOUBLE,
      32, sizeof(double), NULL },
    { "Robot.Position.v", TCL_POSTER_DOUBLE,
      40, sizeof(double), NULL },
    { "Robot.Position.w", TCL_POSTER_DOUBLE,
      48, sizeof(double), NULL },
    { "Robot.PosError.vX(0)", TCL_POSTER_DOUBLE,
      56, sizeof(double), NULL },
    { "Robot.PosError.vX(1)", TCL_POSTER_DOUBLE,
      64, sizeof(double), NULL },
    { "Robot.PosError.vX(2)", TCL_POSTER_DOUBLE,
      72, sizeof(double), NULL },
    { "Robot.PosError.vX(3)", TCL_POSTER_DOUBLE,
      80, sizeof(double), NULL },
    { "Robot.PosError.vX(4)", TCL_POSTER_DOUBLE,
      88, sizeof(double), NULL },
    { "Robot.PosError.vX(5)", TCL_POSTER_DOUBLE,
      96, sizeof(double), NULL },
    { "Robot.PosError.vX(5)", TCL_POSTER_DOUBLE,
      104, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      112, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      120, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      128, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      136, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      144, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      152, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      160, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      168, sizeof(double), NULL },
    { "Robot.PosError.cX(0)", TCL_POSTER_DOUBLE,
      176, sizeof(double), NULL },
    { "Robot.Ref.backFlag", TCL_POSTER_ENUM,
      184, sizeof(int), &locoMotionDirectionEnum },
    { "Robot.Ref.numRef", TCL_POSTER_INT,
      188, sizeof(int), NULL },
    { "Robot.Ref.x", TCL_POSTER_DOUBLE,
      192, sizeof(double), NULL },
    { "Robot.Ref.y", TCL_POSTER_DOUBLE,
      200, sizeof(double), NULL },
    { "Robot.Ref.theta", TCL_POSTER_DOUBLE,
      208, sizeof(double), NULL },
    { "Robot.Ref.v", TCL_POSTER_DOUBLE,
      216, sizeof(double), NULL },
    { "Robot.Ref.w", TCL_POSTER_DOUBLE,
      224, sizeof(double), NULL },
    { "Robot.Ref.vmax", TCL_POSTER_DOUBLE,
      232, sizeof(double), NULL },
    { "Robot.Ref.wmax", TCL_POSTER_DOUBLE,
      240, sizeof(double), NULL },
    { "Robot.Ref.linAccelMax", TCL_POSTER_DOUBLE,
      248, sizeof(double), NULL },
    { "Robot.Ref.angAccelMax", TCL_POSTER_DOUBLE,
      256, sizeof(double), NULL },
    { NULL, TCL_POSTER_NONE, 0, 0, NULL }
};    

/*----------------------------------------------------------------------*/

/**
 ** Init du module Tcl
 **/
int
locoPoster_Init(Tcl_Interp *interp)
{
    int code;
    TCL_POSTER_DESC *d;

    code = Tcl_PkgProvide(interp, "locoPoster", "0.0");
    if (cod != TCL_OK) {
	return code;
    }
    for (d = &tclLocoPosterDesc; d->type != TCL_POSTER_NONE; d++) {
	Tcl_TraceVar(interp, d->name, TCL_TRACE_READS, tclPosterTrace, 
		     (ClientData)d);
    }
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Traites les traces sur les variables d'un poster
 **/
static char *
tclPosterTrace(ClientData clientData, Tcl_Interp *interp,
	       char *name1, char *name2, int flags)
{
    TCL_POSTER_DESC *d = (TCL_POSTER_DESC *)clientData;
    char *buf;
    char tmp[100];

    buf = malloc(d->len);
    
    posterRead(d->posterId, d->offset, d->len, buf);
    switch (d->type) {
      case TCL_POSTER_ENUM:
	Tcl_SetVar2(interp, name1, name2, findEnumString((int *)buf, 
							 d->values),
		    flags & TCL_GLOBAL_ONLY);
	break;
      case TCL_POSTER_INT:
	snprintf(tmp, 99, "%d", *(int *)buf);
	Tcl_SetVar2(interp, name1, name2, tmp,
		    flags & TCL_GLOBAL_ONLY);
	break;
      case TCL_POSTER_FLOAT:
	snprintf(tmp, 99, "%f", *(float *)buf);
	Tcl_SetVar2(interp, name1, name2, tmp,
		    flags & TCL_GLOBAL_ONLY);
	break;
      case TCL_POSTER_DOUBLE:
	snprintf(tmp, 99, "%lf", *(double *)buf);
	Tcl_SetVar2(interp, name1, name2, tmp,
		    flags & TCL_GLOBAL_ONLY);
	break;
      case TCL_POSTER_STRING:
	Tcl_SetVar2(interp, name1, name2, buf, 
		    flags & TCL_GLOBAL_ONLY);
    } /* switch */
    return NULL;
}

	
    
    
