/**
 ** Copyright (c) 1999 Matthieu Herrb
 **
 ** $Source: /cvs/openrobots/grh2/src/robot.c,v $
 ** $Revision: 1.2 $
 ** $Date: 2007/03/28 17:20:52 $
 ** 
 **/

/***
 *** Objet << robot >> pour canvas Tk
 ***
 *** Attention !  Les coordonnes de ces objets passees en parametres
 *** des differentes fonctions sont exprimees directement dans le
 *** repere salle et non en pixels comme pour les objets standard des
 *** canvas Tk.
 *** De plus, pour les robots, il y a une coordonnee theta.
 ***/

#ident "$Id: robot.c,v 1.2 2007/03/28 17:20:52 matthieu Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/* #include <tkInt.h> */
#include <tcl.h>
#include <tk.h>
#include "robot.h"
#include "coord.h"

/*----------------------------------------------------------------------*/

/**
 ** structure interne de l'objet
 **/
typedef struct GRH2_ROBOT_ITEM {
    Tk_Item header;
    double x, y, theta;
    XColor *fgColor;
    XColor *bgColor;
    GC gc;
    char *type;
    int itype;
    double *points;
    int numPoints;
    double r;
} GRH2_ROBOT_ITEM;

/*----------------------------------------------------------------------*/

/**
 ** Options de configuration connues 
 **/
static Tk_CustomOption tagsOption = {Tk_CanvasTagsParseProc,
    Tk_CanvasTagsPrintProc, (ClientData) NULL
};

static Tk_ConfigSpec grh2RobotConfigSpecs[] = {
    { TK_CONFIG_COLOR, "-fill", NULL, NULL,
      "black", Tk_Offset(GRH2_ROBOT_ITEM, fgColor), 0 },
    { TK_CONFIG_STRING, "-type", "type", "Type",
      "h2", Tk_Offset(GRH2_ROBOT_ITEM, type), 0 },
    { TK_CONFIG_CUSTOM, "-tags", NULL, NULL, 
      NULL, 0, TK_CONFIG_NULL_OK, &tagsOption}, 
    { TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0 }
};

/*----------------------------------------------------------------------*/

/**
 ** Polygone decrivant un robot de type h2 
 ***/
double h2[] = {
    -0.60, 0.35,
    0.50, 0.35, 
    0.60, 0.25,
    0.60, -0.25, 
    0.50, -0.35,
    -0.60, -0.35
};

/*----------------------------------------------------------------------*/

/**
 ** Changement de repere robot -> canvas
 **
 ** Attention le repere canvas n'est pas direct, donc la rotation 
 ** est inversee
 **/
static void 
pt_robot_to_salle(GRH2_ROBOT_ITEM *robot, double *pt, double *pres)
{
    double c = cos(robot->theta);
    double s = sin(robot->theta);
    
    pres[0] = robot->x + x_m_to_pixel(pt[0])*c + y_m_to_pixel(pt[1])*s;
    pres[1] = robot->y - x_m_to_pixel(pt[0])*s + y_m_to_pixel(pt[1])*c;
}
/*----------------------------------------------------------------------*/

/**
 ** Calcul du polygone correspondant au robot oriente' 
 ** et mise a jour du bounding box
 **/
static void
updateRobot(Tk_Canvas canvas, GRH2_ROBOT_ITEM *robotPtr)
{
    int i;
    double coord[2];
    Tk_Item *itemPtr = (Tk_Item *)robotPtr;

    switch  (robotPtr->itype) {
      case GRH2_ROBOT_TYPE_H2:
	for (i = 0; i < robotPtr->numPoints* 2; i += 2) {
	    pt_robot_to_salle(robotPtr, &h2[i], robotPtr->points + i);
	    /* Maj du bounding box */
	    coord[0] = robotPtr->points[i];
	    coord[1] = robotPtr->points[i+1];
	    TkIncludePoint((Tk_Item *)robotPtr, coord);
	} /* for */
	/* Il faut etre un peu plus large apparement */
	itemPtr->x1--;
	itemPtr->y1--;
	itemPtr->x2++;
	itemPtr->y2++;
	break;
      case GRH2_ROBOT_TYPE_XR4000:
      case GRH2_ROBOT_TYPE_SCOUT:
	/* Maj du Bounding box directe */
	itemPtr->x1 = robotPtr->x - m_to_pixel(robotPtr->r)
		      - 1;
	itemPtr->y1 = robotPtr->y - m_to_pixel(robotPtr->r)
		      - 1;
	itemPtr->x2 = robotPtr->x + m_to_pixel(robotPtr->r)
		      + 1;
	itemPtr->y2 = robotPtr->y + m_to_pixel(robotPtr->r)
		      + 1;
	break;
      default:
	/* type de robot inconnu */
	break;
	
    } /* switch */
#if 0
    Tk_CanvasEventuallyRedraw(canvas, itemPtr->x1, itemPtr->y1, 
			      itemPtr->x2, itemPtr->y2);
#endif

}
/*----------------------------------------------------------------------*/

/**
 ** Destruction d'un object robot 
 **/
static void
grh2RobotDelete(Tk_Canvas canvas, Tk_Item *itemPtr, Display *display)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;

    if (robotPtr->fgColor != NULL) {
	Tk_FreeColor(robotPtr->fgColor);
    }
    if (robotPtr->bgColor != NULL) {
	Tk_FreeColor(robotPtr->bgColor);
    }
    if (robotPtr->gc != None) {
	Tk_FreeGC(display, robotPtr->gc);
    }
    if (robotPtr->points != NULL) {
	ckfree((char *)robotPtr->points);
    }
}

/*----------------------------------------------------------------------*/

/**
 ** Traitement des options 
 **/
static int
grh2RobotConfig(Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr,
		int argc, char **argv, int flags)
{
    Tk_Window tkwin;
    XGCValues gcValues;
    GC newGC;
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;

    tkwin = Tk_CanvasTkwin(canvas);
    if (Tk_ConfigureWidget(interp, tkwin, grh2RobotConfigSpecs, argc, argv,
			   (char *)itemPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }
    /* Traite Gc */
    gcValues.foreground = robotPtr->fgColor->pixel;
    gcValues.cap_style = CapRound;
    gcValues.join_style = JoinRound;
    newGC = Tk_GetGC(tkwin, GCForeground|GCBackground
		     |GCCapStyle|GCJoinStyle, &gcValues);

    if (robotPtr->gc != None) {
	Tk_FreeGC(Tk_Display(tkwin), robotPtr->gc);
    }
    robotPtr->gc = newGC;

    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Creation d'un nouvel objet de type robot
 **/
static int
grh2RobotCreate(Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr,
		int argc, char **argv)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;

    /* Verification syntaxe */
    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 Tk_PathName(Tk_CanvasTkwin(canvas)), " create ",
			 itemPtr->typePtr->name,
			 " x y theta ?options?\"", NULL);
	return TCL_ERROR;
    }
    /* Recupere parametres */
    if (Tcl_GetDouble(interp, argv[0], &(robotPtr->x)) != TCL_OK) {
	goto error;
    }
    if (Tcl_GetDouble(interp, argv[1], &(robotPtr->y)) != TCL_OK) {
	goto error;
    }
    if (Tcl_GetDouble(interp, argv[2], &(robotPtr->theta)) != TCL_OK) {
	goto error;
    }
    robotPtr->type = NULL;
    robotPtr->fgColor = NULL;
    robotPtr->bgColor = NULL;
    robotPtr->gc = NULL;
    robotPtr->points = NULL;

    /* Parse les options */
    if (grh2RobotConfig(interp, canvas, itemPtr, argc - 3, argv + 3, 0) 
	!= TCL_OK) {
	goto error;
    }
    /* Recupere le type */
    if (strncmp(robotPtr->type, "h2", 2) == 0) {
	robotPtr->itype = GRH2_ROBOT_TYPE_H2;
	robotPtr->points = (double *)ckalloc(12*sizeof(double));
	robotPtr->numPoints = 6;
	robotPtr->r = 0.0;
    } else if (strncmp(robotPtr->type, "xr4000", 6) == 0) {
	robotPtr->itype = GRH2_ROBOT_TYPE_XR4000;
	robotPtr->numPoints = 0;
	robotPtr->r = XR4000_RADIUS;
    } else if (strncmp(robotPtr->type, "scout", 5) == 0) {
	robotPtr->itype = GRH2_ROBOT_TYPE_SCOUT;
	robotPtr->numPoints = 0;
	robotPtr->r = SCOUT_RADIUS;
    } else {
	Tcl_AppendResult(interp, "unknown robot type: ", 
			 robotPtr->type, NULL);
	goto error;
    }
    /* Calcul de la bounding box */
    updateRobot(canvas, robotPtr);
    return TCL_OK;

 error:
    /* On arrive ici en cas d'erreur */
    /* grh2RobotDelete(canvas, itemPtr, Tk_Display(Tk_CanvasTkwin(canvas))); */
    return TCL_ERROR;
}

/*----------------------------------------------------------------------*/

/**
 ** Modification des coordonnees d'un robot
 **/
static int
grh2RobotCoord(Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr,
	       int argc, char **argv)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;
    char buffer[TCL_DOUBLE_SPACE];

    if (argc == 0) {
	/* Retourne coordonnees */
	Tcl_PrintDouble(interp, robotPtr->x, buffer);
	Tcl_AppendElement(interp, buffer);
	Tcl_PrintDouble(interp, robotPtr->y, buffer);
	Tcl_AppendElement(interp, buffer);
	Tcl_PrintDouble(interp, robotPtr->theta, buffer);
	Tcl_AppendElement(interp, buffer);
    } else if (argc != 3) {
	Tcl_AppendResult(interp, "too few coordinates for robot: "
			 "must have at least 3", NULL);
	return TCL_ERROR;
    } else {
	if (Tcl_GetDouble(interp, argv[0], &(robotPtr->x)) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (Tcl_GetDouble(interp, argv[1], &(robotPtr->y)) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (Tcl_GetDouble(interp, argv[2], &(robotPtr->theta)) != TCL_OK) {
	    return TCL_ERROR;
	}
	updateRobot(canvas, robotPtr);
    }
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Affichage 
 **/
static void
grh2RobotDisplay(Tk_Canvas canvas, Tk_Item *itemPtr, Display *display,
		 Drawable dst, int x, int y, int width, int height)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;
    int i;
    static XPoint pix_robot[7];
    double *coordPtr;

    if (robotPtr->gc == None) {
	return;
    }
    switch (robotPtr->itype) {
      case GRH2_ROBOT_TYPE_H2:
	/* Calcule coordonnes canvas */
	for (i = 0, coordPtr = robotPtr->points; i < robotPtr->numPoints; 
	     i++, coordPtr += 2) {
	    Tk_CanvasDrawableCoords(canvas, 
				    coordPtr[0], coordPtr[1],
				    &(pix_robot[i].x), &pix_robot[i].y);
	} /* for */
	/* Ferme le contour */
	pix_robot[i].x = pix_robot[0].x;
	pix_robot[i].y = pix_robot[0].y;
	/* Dessin */
	XDrawLines(display, dst, robotPtr->gc, pix_robot,
		   robotPtr->numPoints+1, CoordModeOrigin);
	break;
      case GRH2_ROBOT_TYPE_XR4000:
      case GRH2_ROBOT_TYPE_SCOUT:
	/* Calcule coordonnees canvas */
	Tk_CanvasDrawableCoords(canvas, robotPtr->x, robotPtr->y,
				&pix_robot[0].x, &pix_robot[0].y);
	Tk_CanvasDrawableCoords(canvas, robotPtr->x + 
				m_to_pixel(robotPtr->r)*cos(robotPtr->theta),
				robotPtr->y - 
				m_to_pixel(robotPtr->r)*sin(robotPtr->theta),
				&pix_robot[1].x, &pix_robot[1].y);
	/* Dessin orientation */
	XDrawLine(display, dst, robotPtr->gc, pix_robot[0].x, pix_robot[0].y,
		  pix_robot[1].x, pix_robot[1].y);
	/* Coordonnes canvas pour le cercle */
	Tk_CanvasDrawableCoords(canvas, 
				robotPtr->x - m_to_pixel(robotPtr->r),
				robotPtr->y - m_to_pixel(robotPtr->r),
				&pix_robot[0].x, &pix_robot[0].y);
	/* Dessin du cercle */
	XDrawArc(display, dst, robotPtr->gc, pix_robot[0].x, pix_robot[0].y,
		 2*m_to_pixel(robotPtr->r), 2*m_to_pixel(robotPtr->r),
		 0, 360*64 - 1);
		  
	break;
	
    } /* case */
}

/*----------------------------------------------------------------------*/

/**
 ** Calcul de la distance d'un point au robot 
 **/
static double
grh2RobotPoint(Tk_Canvas canvas, Tk_Item *itemPtr, double *pointPtr)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;
    double d;

    switch (robotPtr->itype) {
      case GRH2_ROBOT_TYPE_H2:
	return TkPolygonToPoint(robotPtr->points, 
				robotPtr->numPoints, 
				pointPtr);
      case GRH2_ROBOT_TYPE_XR4000:
	d = hypot(pointPtr[0] - robotPtr->x, pointPtr[1] - robotPtr->y) 
	    - m_to_pixel(robotPtr->r);
	if (d <= 0) {
	    return 0.0;
	} else {
	    /* Transforme le resultat en pixels */
	    return d;
	}
    }
    return 0.0;
}

/*----------------------------------------------------------------------*/

/**
 ** Calcule recouvrement par un rectangle 
 **
 ** retourne 1 si entierement inclus 
 **         -1 si entierement a l'exterieur
 **          0 si recouvrement partiel
 **/
static int
grh2RobotArea(Tk_Canvas canvas, Tk_Item *itemPtr, double *rectPtr)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;
    double r = m_to_pixel(robotPtr->r);

    switch (robotPtr->itype) {
      case GRH2_ROBOT_TYPE_H2:
	return TkPolygonToArea(robotPtr->points, 
			       robotPtr->numPoints, rectPtr);
      case GRH2_ROBOT_TYPE_XR4000:
	if (robotPtr->x > rectPtr[0] + r &&
	    robotPtr->x < rectPtr[2] - r &&
	    robotPtr->y > rectPtr[1] + r &&
	    robotPtr->y < rectPtr[3] - r) {
	    /* entierement a l'intrieur */
	    return 1;
	} else if ((robotPtr->x < rectPtr[0] - r ||
		    robotPtr->x > rectPtr[2] + r) &&
		   (robotPtr->y < rectPtr[1] - r || 
		    robotPtr->y > rectPtr[3] + r)) {
	    /* entierement a l'exterieur */
	    return -1;
	} else {
	    return 0;
	}
    } /* switch */
    return -1;
}

/*----------------------------------------------------------------------*/

/**
 ** Produit PostScript
 **/
static int
grh2RobotPostscript(Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr,
		    int prepass)
{
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Changement d'echelle 
 **/
static void
grh2RobotScale(Tk_Canvas canvas, Tk_Item *itemPtr, 
	       double originX, double originY, double scaleX, double scaleY)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;
    
    robotPtr->x = originX + scaleX*(robotPtr->x - originX);
    robotPtr->y = originY + scaleY*(robotPtr->y - originY);
    /* Recalcule la bounding box */
    updateRobot(canvas, robotPtr);
}

/*----------------------------------------------------------------------*/

/**
 ** Translation
 **/
static void
grh2RobotTranslate(Tk_Canvas canvas, Tk_Item *itemPtr, 
		   double deltaX, double deltaY)
{
    GRH2_ROBOT_ITEM *robotPtr = (GRH2_ROBOT_ITEM *)itemPtr;

    robotPtr->x += deltaX;
    robotPtr->y += deltaY;
    /* Recalcule la bounding box */
    updateRobot(canvas, robotPtr);
}

/*----------------------------------------------------------------------*/

/**
 ** Structure definissant le nouvel object
 **/
Tk_ItemType robot = {
    "robot",
    sizeof(GRH2_ROBOT_ITEM),
    grh2RobotCreate,
    grh2RobotConfigSpecs,
    grh2RobotConfig,
    grh2RobotCoord,
    grh2RobotDelete,
    grh2RobotDisplay,
    0,
    grh2RobotPoint,
    grh2RobotArea,
    grh2RobotPostscript,
    grh2RobotScale,
    grh2RobotTranslate,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};

